import sys
import traceback
import mosamatic2.constants as constants
from PySide6 import QtWidgets
from mosamatic2.ui.settings import Settings
from mosamatic2.core.managers.logmanager import LogManager
from mosamatic2.ui.widgets.splashscreen import SplashScreen

LOG = LogManager()


def run_tests():
    return 'PASSED'


def main():
    settings = Settings()
    application_name = settings.get(constants.MOSAMATIC2_WINDOW_TITLE)
    QtWidgets.QApplication.setApplicationName(application_name)
    app = QtWidgets.QApplication(sys.argv)
    app.setApplicationName(application_name)
    try:
        splash = SplashScreen()
        splash.show()
        sys.exit(app.exec())
    except Exception as e:
        LOG.error(str(e))
        LOG.error(traceback.format_exc())
        sys.exit(1)

if __name__ == '__main__':
    main()


import os
from mosamatic2.core.pipelines.pipeline import Pipeline
from mosamatic2.core.managers.logmanager import LogManager
from mosamatic2.core.utils import is_docker_running, to_unix_path

LOG = LogManager()


class DefaultDockerPipeline(Pipeline):
    INPUTS = [
        'images',
        'model_files',
    ]
    PARAMS = [
        'target_size',
        'file_type',
        'fig_width',
        'fig_height',
        'model_type',
        'model_version',
        'version',
    ]
    def __init__(self, inputs, params, output, overwrite):
        super(DefaultDockerPipeline, self).__init__(inputs, params, output, overwrite)

    def run(self):
        assert is_docker_running()
        docker_script = 'docker run --rm ' + \
            '-v "{}":/data/images '.format(to_unix_path(self.input('images'))) + \
            '-v "{}":/data/model_files '.format(to_unix_path(self.input('model_files'))) + \
            '-v "{}":/data/output '.format(to_unix_path(self.output())) + \
            'brecheisen/mosamatic2-cli:{} defaultpipeline '.format(self.param('version')) + \
            '--images /data/images --model_files /data/model_files --output /data/output --overwrite true'
        LOG.info(f'Running Docker script: {docker_script}')
        os.system(docker_script)


from mosamatic2.core.pipelines.defaultpipeline.defaultpipeline import DefaultPipeline
from mosamatic2.core.pipelines.defaultdockerpipeline.defaultdockerpipeline import DefaultDockerPipeline
from mosamatic2.core.pipelines.boadockerpipeline.boadockerpipeline import BoaDockerPipeline


import os
import shutil
from mosamatic2.core.managers.logmanager import LogManager
from mosamatic2.core.utils import create_name_with_timestamp, mosamatic_output_dir

LOG = LogManager()


class Task:
    INPUTS = []
    PARAMS = []
    OUTPUT = 'output'

    def __init__(self, inputs, params, output, overwrite=True):
        self._inputs = inputs
        self._params = params
        self._output = os.path.join(output, self.__class__.__name__.lower())
        self._overwrite = overwrite
        if self._overwrite and os.path.isdir(self._output):
            shutil.rmtree(self._output)
        os.makedirs(self._output, exist_ok=self._overwrite)
        # Check that the inputs match specification and type
        assert isinstance(self._inputs, dict)
        assert len(self._inputs.keys()) == len(self.__class__.INPUTS)
        for k, v in self._inputs.items():
            assert k in self.__class__.INPUTS
            assert isinstance(v, str)
        # Check that param names match specification (if not None)
        if self._params:
            assert len(self._params.keys()) == len(self.__class__.PARAMS)
            for k in self._params.keys():
                assert k in self.__class__.PARAMS

    def input(self, name):
        return self._inputs[name]
    
    def param(self, name):
        return self._params[name]
    
    def output(self):
        return self._output
    
    def overwrite(self):
        return self._overwrite
    
    def set_progress(self, step, nr_steps):
        LOG.info(f'[{self.__class__.__name__}] step {step} from {nr_steps}')
    
    def run(self):
        raise NotImplementedError()


import os
from mosamatic2.core.pipelines.pipeline import Pipeline
from mosamatic2.core.tasks import Dicom2NiftiTask
from mosamatic2.core.managers.logmanager import LogManager
from mosamatic2.core.utils import (
    is_docker_running, 
    to_unix_path,
    current_time_in_seconds,
    elapsed_time_in_seconds,
    duration,
)

LOG = LogManager()


class BoaDockerPipeline(Pipeline):
    INPUTS = ['scans']
    PARAMS = []

    def __init__(self, inputs, params, output, overwrite):
        super(BoaDockerPipeline, self).__init__(inputs, params, output, overwrite)

    def load_nifti_files(self):
        nifti_files_dir = os.path.join(self.output(), 'dicom2niftitask')
        nifti_files = []
        for f in os.listdir(nifti_files_dir):
            f_path = os.path.join(nifti_files_dir, f)
            if f.endswith('.nii') or f.endswith('.nii.gz'):
                nifti_files.append(f_path)
        return nifti_files

    def run(self):
        assert is_docker_running()
        # First convert DICOM series to NIFTI format
        dicom2nifti_task = Dicom2NiftiTask(
            inputs={'scans': self.input('scans')},
            params={'compressed': True},
            output=self.output(),
            overwrite=self.overwrite(),
        )
        dicom2nifti_task.run()
        # Load NIFTI file paths
        nifti_files = self.load_nifti_files()
        workspaces_dir = to_unix_path(os.path.join(self.output(), 'workspaces'))
        os.makedirs(workspaces_dir, exist_ok=True)
        weights_dir = to_unix_path(os.path.join(self.output(), 'weights'))
        start_time = current_time_in_seconds()
        for f in nifti_files:
            start_time_f = current_time_in_seconds()
            f = to_unix_path(f)
            dir_name = os.path.split(f)[1][:-7]
            workspace = to_unix_path(os.path.join(workspaces_dir, dir_name))
            # Build Docker script
            docker_script = 'docker run --rm ' + \
                '-v "{}":/image.nii.gz '.format(f) + \
                '-v "{}":/workspace '.format(workspace) + \
                '-v "{}":/app/weights '.format(weights_dir) + \
                '--gpus all ' + \
                '--network host ' + \
                '--shm-size=8g --ulimit memlock=-1 --ulimit stack=67108864 ' + \
                '--entrypoint /bin/sh ' + \
                'shipai/boa-cli -c ' + \
                '"python body_organ_analysis --input-image /image.nii.gz --output-dir /workspace/ --models all --verbose"'
            LOG.info(f'Running BOA Docker script: {docker_script}')
            # Run Docker script
            os.system(docker_script)
            elapsed_f = elapsed_time_in_seconds(start_time_f)
            LOG.info(f'Elapsed {dir_name}: {duration(elapsed_f)}')
        elapsed_total = elapsed_time_in_seconds(start_time)
        LOG.info(f'Elapsed total: {duration(elapsed_total)}')


from PySide6.QtWidgets import (
    QDialog,
)


class Dialog(QDialog):
    def __init__(self, parent=None):
        super(Dialog, self).__init__(parent)
        self.setFixedWidth(400)

    def clear(self):
        raise NotImplementedError()

    def showEvent(self, arg__1):
        self.clear()
        return super().showEvent(arg__1)


import click
from mosamatic2.commands import (
    calculatescores,
    rescaledicomimages,
    segmentmusclefatl3tensorflow,
    createpngsfromsegmentations,
    dicom2nifti,
    selectslicefromscans,
    createdicomsummary,
    defaultpipeline,
    defaultdockerpipeline,
    boadockerpipeline,
)
from mosamatic2.core.utils import show_doc_command


class CustomHelpGroup(click.Group):
    def format_commands(self, ctx, formatter):
        commands = self.list_commands(ctx)
        with formatter.section('Commands'):
            for command_name in commands:
                command = self.get_command(ctx, command_name)
                if command is None or command.hidden:
                    continue
                help_text = command.get_short_help_str()
                formatter.write_text(f'{command_name:15} {help_text}')


@click.group(cls=CustomHelpGroup)
def main():
    pass


main.add_command(calculatescores.calculatescores)
main.add_command(rescaledicomimages.rescaledicomimages)
main.add_command(segmentmusclefatl3tensorflow.segmentmusclefatl3tensorflow)
main.add_command(createpngsfromsegmentations.createpngsfromsegmentations)
main.add_command(dicom2nifti.dicom2nifti)
main.add_command(selectslicefromscans.selectslicefromscans)
main.add_command(createdicomsummary.createdicomsummary)
main.add_command(defaultpipeline.defaultpipeline)
main.add_command(defaultdockerpipeline.defaultdockerpipeline)
main.add_command(boadockerpipeline.boadockerpipeline)
main.add_command(show_doc_command(main)) # Special command to show long description for command


from mosamatic2.core.data.filedata import FileData
from mosamatic2.core.utils import (
    is_dicom,
    load_dicom,
    is_jpeg2000_compressed,
)
from mosamatic2.core.managers.logmanager import LogManager

LOG = LogManager()


class DicomImage(FileData):
    def load(self):
        if self.path():
            p = load_dicom(self.path())
            if p:
                if is_jpeg2000_compressed(p):
                    p.decompress()
                self.set_object(p)
                return True
            # if is_dicom(self.path()):
            #     p = load_dicom(self.path())
            #     if is_jpeg2000_compressed(p):
            #         p.decompress()
            #     self.set_object(p)
            #     return True
        return False


import click

from mosamatic2.core.tasks import RescaleDicomImagesTask


@click.command(help='Rescale DICOM files to target size')
@click.option(
    '--images', 
    required=True, 
    type=click.Path(exists=True), 
    help='Input directory with images'
)
@click.option(
    '--output', 
    required=True, 
    type=click.Path(), 
    help='Output directory'
)
@click.option(
    '--target_size', 
    default=512,
    help='Target size of rescaled images (default: 512)'
)
@click.option(
    '--overwrite', 
    type=click.BOOL, 
    default=False, 
    help='Overwrite [true|false]'
)
def rescaledicomimages(images, output, target_size, overwrite):
    """
    Rescales DICOM images to given (square) target size

    Parameters
    ----------
    --images : str
        Directory with input DICOM images

    --output : str
        Path to output directory

    --target_size : int
        Target size for rescaled images (default: 512)
    
    --overwrite : bool
        Overwrite contents output directory: [true|false]
    """
    task = RescaleDicomImagesTask(
        inputs={'images': images}, 
        params={'target_size': target_size}, 
        output=output, 
        overwrite=overwrite
    )
    task.run()


import click

from mosamatic2.core.pipelines import DefaultDockerPipeline


@click.command(help='Runs default L3 analysis pipeline through Docker')
@click.option(
    '--images', 
    required=True, 
    type=click.Path(exists=True), 
    help='Directory with L3 images (no spaces allowed)',
)
@click.option(
    '--model_files', 
    required=True, 
    type=click.Path(), 
    help='Input directory with AI model files (no spaces allowed)'
)
@click.option(
    '--output', 
    required=True, 
    type=click.Path(), 
    help='Output directory (no spaces allowed)'
)
@click.option(
    '--overwrite', 
    type=click.BOOL, 
    default=False, 
    help='Overwrite [true|false]'
)
def defaultdockerpipeline(images, model_files, output, overwrite):
    """
    Runs default L3 analysis pipeline through Docker
    
    Parameters
    ----------
    --images : str
        Directory with L3 images. Warning: This directory path cannot contain any spaces!
        Docker has issues working with that.

    --model_files : str
        Directory with AI model files. This directory can ONLY contain the following
        files:

        (1) contour_model-1.0.zip
        (2) model-1.0.zip
        (3) params-1.0.json

        These files can be downloaded from:
        https://mosamatic.rbeesoft.nl/wordpress/mosamatic/installation/

        Warning: This directory path cannot contain any spaces!

    --output : str
        Path to output directory (no spaces!)

    --overwrite : bool
        Overwrite contents output directory [true|false]
    """
    pipeline = DefaultDockerPipeline(
        inputs={'images': images, 'model_files': model_files},
        params={
            'target_size': 512,
            'model_type': 'tensorflow',
            'model_version': 1.0,
            'file_type': 'npy',
            'fig_width': 10,
            'fig_height': 10,
        },
        output=output,
        overwrite=overwrite,
    )
    pipeline.run()


import os

from mosamatic2.core.pipelines.pipeline import Pipeline
from mosamatic2.core.tasks import (
    RescaleDicomImagesTask, 
    SegmentMuscleFatL3TensorFlowTask,
    SegmentMuscleFatL3TensorFlowTask,
    CreatePngsFromSegmentationsTask,
    CalculateScoresTask,
)
from mosamatic2.core.managers.logmanager import LogManager

LOG = LogManager()


class DefaultPipeline(Pipeline):
    INPUTS = [
        'images',
        'model_files',
    ]
    PARAMS = [
        'target_size',
        'file_type',
        'fig_width',
        'fig_height',
        'model_type',
        'model_version',
    ]
    def __init__(self, inputs, params, output, overwrite):
        super(DefaultPipeline, self).__init__(inputs, params, output, overwrite)
        LOG.info('Found {} images to process'.format(len(os.listdir(self.input('images')))))
        model_type = self.param('model_type')
        # segmentation_task_class = SegmentMuscleFatL3Task if model_type == 'pytorch' else SegmentMuscleFatL3TensorFlowTask
        segmentation_task_class = SegmentMuscleFatL3TensorFlowTask
        self.add_task(
            RescaleDicomImagesTask(
                inputs={'images': self.input('images')},
                params={'target_size': self.param('target_size')},
                output=self.output(),
                overwrite=self.overwrite(),
            )
        )
        self.add_task(
            segmentation_task_class(
                inputs={
                    'images': os.path.join(self.output(), 'rescaledicomimagestask'),
                    'model_files': self.input('model_files'),
                },
                params={'model_version': self.param('model_version')},
                output=self.output(),
                overwrite=self.overwrite(),
            )
        )
        self.add_task(
            CalculateScoresTask(
                inputs={
                    'images': os.path.join(self.output(), 'rescaledicomimagestask'),
                    'segmentations': os.path.join(
                        self.output(),
                        'segmentmusclefatl3pytorchtask' if model_type == 'pytorch' else 'segmentmusclefatl3tensorflowtask',
                    )
                },
                params={'file_type': self.param('file_type')},
                output=self.output(),
                overwrite=self.overwrite(),
            )
        )
        self.add_task(
            CreatePngsFromSegmentationsTask(
                inputs={
                    'segmentations': os.path.join(
                        self.output(),
                        'segmentmusclefatl3pytorchtask' if model_type == 'pytorch' else 'segmentmusclefatl3tensorflowtask',
                    )
                },
                params={
                    'fig_width': self.param('fig_width'),
                    'fig_height': self.param('fig_height'),
                },
                output=self.output(),
                overwrite=self.overwrite(),
            )
        )


import vtk


class CustomInteractorStyle(vtk.vtkInteractorStyleImage):
    def __init__(self, image_data, slice_mapper, status_actor, slice_obj, orientation="axial"):
        super(CustomInteractorStyle, self).__init__()
        self.AddObserver("MouseWheelForwardEvent", self.move_slice_forward)
        self.AddObserver("MouseWheelBackwardEvent", self.move_slice_backward)
        self.AddObserver("MouseMoveEvent", self.update_overlay, 1.0)
        self.AddObserver("KeyPressEvent", self.key_press_event)

        self.image_data = image_data
        self.slice_mapper = slice_mapper
        self.status_actor = status_actor
        self.slice_obj = slice_obj
        self.orientation = orientation

        xmin, xmax, ymin, ymax, zmin, zmax = image_data.GetExtent()
        if orientation == "axial":
            self.min_slice, self.max_slice = zmin, zmax
        elif orientation == "sagittal":
            self.min_slice, self.max_slice = xmin, xmax
        elif orientation == "coronal":
            self.min_slice, self.max_slice = ymin, ymax
        else:
            raise ValueError(f"Unknown orientation: {orientation}")

        self.slice = (self.min_slice + self.max_slice) // 2
        self.slice_mapper.SetSliceNumber(self.slice)
        self.update_status_message()

    def update_status_message(self):
        window = int(self.slice_obj.GetProperty().GetColorWindow())
        level = int(self.slice_obj.GetProperty().GetColorLevel())
        message = f'Slice {self.slice + 1}/{self.max_slice + 1} | W: {window} L: {level}'
        self.status_actor.GetMapper().SetInput(message)

    def move_slice_forward(self, obj, event):
        if self.slice < self.max_slice:
            self.slice += 1
            self.slice_mapper.SetSliceNumber(self.slice)
            self.update_status_message()
            self.GetInteractor().GetRenderWindow().Render()

    def move_slice_backward(self, obj, event):
        if self.slice > self.min_slice:
            self.slice -= 1
            self.slice_mapper.SetSliceNumber(self.slice)
            self.update_status_message()
            self.GetInteractor().GetRenderWindow().Render()

    def key_press_event(self, obj, event):
        key = self.GetInteractor().GetKeySym()
        if key == "Up":
            self.move_slice_forward(obj, event)
        elif key == "Down":
            self.move_slice_backward(obj, event)

    def update_overlay(self, obj, event):
        super(CustomInteractorStyle, self).OnMouseMove()
        self.update_status_message()
        self.GetInteractor().GetRenderWindow().Render()


import os
from mosamatic2.core.tasks.task import Task
from mosamatic2.core.managers.logmanager import LogManager
from mosamatic2.core.utils import (
    convert_numpy_array_to_png_image, 
    AlbertaColorMap, 
    load_numpy_array,
)

LOG = LogManager()


class CreatePngsFromSegmentationsTask(Task):
    INPUTS = ['segmentations']
    PARAMS = [
        'fig_width', 
        'fig_height'
    ]

    def __init__(self, inputs, params, output, overwrite=True):
        super(CreatePngsFromSegmentationsTask, self).__init__(inputs, params, output, overwrite)

    def load_segmentations(self):
        segmentations = []
        for f in os.listdir(self.input('segmentations')):
            f_path = os.path.join(self.input('segmentations'), f)
            if f.endswith('.seg.npy'):
                segmentations.append(f_path)
        if len(segmentations) == 0:
            raise RuntimeError('Input directory has no segmentation files')
        return segmentations

    def run(self):
        segmnentations = self.load_segmentations()
        nr_steps = len(segmnentations)
        for step in range(nr_steps):
            source = segmnentations[step]
            source_name = os.path.split(source)[1]
            source_image = load_numpy_array(source)
            if source_image is not None:
                png_file_name = source_name + '.png'
                convert_numpy_array_to_png_image(
                    source_image, 
                    self.output(),
                    AlbertaColorMap(), 
                    png_file_name,
                    fig_width=int(self.param('fig_width')),
                    fig_height=int(self.param('fig_height')),
                )
            else:
                LOG.warning(f'File {source} is not a valid NumPy array file')
            self.set_progress(step, nr_steps)



from PySide6.QtWidgets import QStackedWidget


class StackedPanel(QStackedWidget):
    def __init__(self):
        super(StackedPanel, self).__init__()
        self._current_panel_name = None
        self._panels = {}

    def add_panel(self, panel, name):
        if name not in self._panels.keys():
            self._panels[name] = panel
            self._current_panel_name = name
            self.addWidget(panel)

    def current_panel_name(self):
        return self._current_panel_name

    def switch_to(self, name):
        panel = self._panels[name]
        self._current_panel_name = name
        self.setCurrentWidget(panel)